"""
Fritter, the Frame-Rate IndependenT TimEr tRee.

This is a library for scheduling timed events in the very near or very far
future.
"""

from . import (
    heap,
    tree,
    persistent,
    boundaries,
    drivers,
    scheduler,
    repeat,
)

__version__ = "0.1.0"
"The current version of the Fritter library."

__all__ = [
    "heap",
    "tree",
    "persistent",
    "boundaries",
    "drivers",
    "scheduler",
    "repeat",
]



# -*- test-case-name: fritter.test.test_scheduler -*-
"""
A L{Scheduler} is the core interface of Fritter; a collection of timed calls
scheduled by L{callAt <Scheduler.callAt>} connected to a L{TimeDriver} that
causes them to actually be called.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from itertools import count
from typing import Callable, Generic, overload

from .boundaries import (
    IDT,
    PriorityQueue,
    ScheduledCall,
    ScheduledState,
    Scheduler,
    TimeDriver,
    WhatT,
    WhenT,
)
from .heap import Heap


@dataclass(eq=True, order=True)
class ConcreteScheduledCall(Generic[WhenT, WhatT, IDT]):
    """
    A handle to a call that has been scheduled.
    """

    _when: WhenT = field(compare=True)
    _what: WhatT | None = field(compare=False)
    _id: IDT = field(compare=True)

    _called: bool = field(compare=False)
    _cancelled: bool = field(compare=False)
    _canceller: (
        Callable[[ConcreteScheduledCall[WhenT, WhatT, IDT]], None] | None
    ) = field(compare=False)

    def _call(self) -> None:
        """
        Invoke the callable and adjust the state.
        """
        assert self._what is not None, "ScheduledCall invoked twice."
        self._called = True
        try:
            self._what()
        finally:
            self._what = None

    @property
    def id(self) -> IDT:
        """
        Return a unique identifier for this scheduled call.
        """
        return self._id

    @property
    def when(self) -> WhenT:
        """
        Return the original time at which the call will be scheduled.
        """
        return self._when

    @property
    def what(self) -> WhatT | None:
        """
        If this has not been called or cancelled, return the original callable
        that was scheduled.

        @note: To break cycles, this will only have a non-C{None} value when in
            L{ScheduledState.pending}.
        """
        return self._what

    @property
    def state(self) -> ScheduledState:
        """
        Is this call still waiting to be called, or has it been called or
        cancelled?
        """
        if self._cancelled:
            return ScheduledState.cancelled
        if self._called:
            return ScheduledState.called
        assert self._what is not None
        return ScheduledState.pending

    def cancel(self) -> None:
        """
        Cancel this L{ScheduledCall}, making it so that it will not be invoked
        in the future.  If the work described by C{when} has already been
        called, or this call has already been cancelled, do nothing.
        """
        if self._canceller is None:
            return
        self._cancelled = True
        try:
            self._canceller(self)
        finally:
            self._canceller = None


@dataclass
class _PriorityQueueBackedSchedulerImpl(Generic[WhenT, WhatT, IDT]):
    """
    A L{Scheduler} allows for scheduling work (of the type C{WhatT}, which must
    be at least a 0-argument None-returning callable) at a given time
    (C{WhenT}, which much be sortable as a L{PriorityComparable}).

    @ivar driver: The L{TimeDriver} that this L{Scheduler} will use.
    """

    driver: TimeDriver[WhenT]
    _newID: Callable[[], IDT]
    _q: PriorityQueue[ConcreteScheduledCall[WhenT, WhatT, IDT]]
    _maxWorkBatch: int = 0xFF

    def now(self) -> WhenT:
        """
        Relay C{now} to our L{TimeDriver}.
        """
        return self.driver.now()

    def callAt(
        self, when: WhenT, what: WhatT
    ) -> ScheduledCall[WhenT, WhatT, IDT]:
        """
        Call C{what} at the time C{when} according to the L{TimeDriver}
        associated with this L{Scheduler}.

        @return: a L{ScheduledCall} that describes the pending call and allows
            for cancelling it.
        """

        def advanceToNow() -> None:
            timestamp = self.driver.now()
            workPerformed = 0
            while (
                (each := self._q.peek()) is not None
                and each._when <= timestamp
                and workPerformed < self._maxWorkBatch
            ):
                popped = self._q.get()
                assert popped is each
                each._call()
                workPerformed += 1
            upNext = self._q.peek()
            if upNext is not None:
                self.driver.reschedule(upNext._when, advanceToNow)

        def _cancelCall(
            toRemove: ConcreteScheduledCall[WhenT, WhatT, IDT],
        ) -> None:
            old = self._q.peek()
            self._q.remove(toRemove)
            new = self._q.peek()
            if new is None:
                self.driver.unschedule()
            elif old is None or new is not old:
                self.driver.reschedule(new._when, advanceToNow)

        previously = self._q.peek()
        call = ConcreteScheduledCall(
            when, what, self._newID(), False, False, _cancelCall
        )
        self._q.add(call)
        currently = self._q.peek()
        # We just added a thing it can't be None even though peek has that
        # signature
        assert currently is not None
        if previously is None or previously._when != currently._when:
            self.driver.reschedule(currently._when, advanceToNow)
        return call


_TypeCheck: type[Scheduler[float, Callable[[], None], int]] = (
    _PriorityQueueBackedSchedulerImpl
)


@overload
def schedulerFromDriver(
    driver: TimeDriver[WhenT],
    nextID: Callable[[], IDT],
    queue: (
        PriorityQueue[ConcreteScheduledCall[WhenT, WhatT, IDT]] | None
    ) = None,
) -> Scheduler[WhenT, WhatT, IDT]: ...


@overload
def schedulerFromDriver(
    driver: TimeDriver[WhenT],
    *,
    queue: (
        PriorityQueue[ConcreteScheduledCall[WhenT, WhatT, int]] | None
    ) = None,
) -> Scheduler[WhenT, WhatT, int]: ...


def schedulerFromDriver(
    driver: TimeDriver[WhenT],
    nextID: Callable[[], IDT] | None = None,
    queue: (
        PriorityQueue[ConcreteScheduledCall[WhenT, WhatT, IDT]] | None
    ) = None,
) -> Scheduler[WhenT, WhatT, IDT]:
    """
    Create a new in-memory scheduler.

    @param driver: The L{TimeDriver} to use for the new scheduler.

    @param queue: If desired, a custom L{PriorityQueue} implementation.  By
        default, a new L{Heap} will be used.

    @param nextID: A callable that will generate new opaque IDs.  By default,
        sequential integers will be used.
    """
    if nextID is None:
        nextCounter = count().__next__
        # I really just want https://peps.python.org/pep-0696/ here (the
        # default type for "IDT" is "int", "nextCounter" is Callable[, int]
        # where we want a Callable[, IDT] and there's no way to tell mypy those
        # are the same but only in this case), but we'll have to wait for
        # Python 3.13.
        nextID = nextCounter  # type:ignore[assignment]
        assert (
            nextID is not None
        ), "itertools.count.__next__ just isn't None, but mypy can't tell"
    return _PriorityQueueBackedSchedulerImpl[WhenT, WhatT, IDT](
        driver, nextID, Heap() if queue is None else queue
    )


__all__ = [
    "ConcreteScheduledCall",
    "schedulerFromDriver",
]



"""
Implementation of a L{TimeDriver} and a L{Scale} to convert floating-point
POSIX timestamps into timezone-aware datetimes.

@note: Although at runtime this module uses L{datetime.datetime} objects, its
    type hints use the U{datetype <https://pypi.org/project/datetype/>} library
    to ensure that all values have the correct type of tzinfo.  This driver
    will I{not} work with arbitrary L{datetime.timezone} zones, I{only}
    L{zoneinfo.ZoneInfo} zones, as IANA timezone identifiers are required for
    reliable long-term serialization across DST boundaries.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Callable
from zoneinfo import ZoneInfo

from datetype import DateTime

from ..boundaries import (
    TimeDriver,
    Scale,
)


__all__ = [
    "guessLocalZone",
    "DateScale",
    "DateTimeDriver",
]

_PS_TZ_CMD = """\
powershell \
[\
Windows.Globalization.Calendar,\
Windows.Globalization,\
ContentType=WindowsRuntime\
]\
::New().GetTimeZone()\
"""

_guessedZone: ZoneInfo | None = None


def guessLocalZone() -> ZoneInfo:
    """
    Attempt to determine the IANA timezone identifier for the local system
    using a variety of heuristics.
    """
    global _guessedZone
    if _guessedZone is not None:
        return _guessedZone
    from os import name, readlink

    if name == "nt":
        from os import popen

        ianaID = popen(_PS_TZ_CMD).read().strip()
    else:
        path = readlink("/etc/localtime").split("/")
        ianaID = "/".join(path[path.index("zoneinfo") + 1 :])
    _guessedZone = ZoneInfo(ianaID)
    return _guessedZone


@dataclass
class DateScale:
    """
    A L{fritter.boundaries.Scale} that can scale between a trunk scheduler that
    uses physical time (floats) and a branch scheduler that uses civil time
    (dates and wall-clock times).
    """

    _zone: ZoneInfo
    _onlyPauseShift: bool = True

    def up(self, offset: float, time: DateTime[ZoneInfo]) -> float:
        """
        Translate C{time} from the branch time scale into the trunk time scale.
        """
        return time.timestamp() + offset

    def down(self, offset: float, time: float) -> DateTime[ZoneInfo]:
        """
        Translate C{time} from the trunk time scale into the branch time scale.
        """
        return DateTime.fromtimestamp(time - offset, self._zone)

    def shift(
        self, pauseTime: DateTime[ZoneInfo] | None, currentTime: float
    ) -> float:
        """
        Shift the current scale forward to incorporate pause breaks.
        """
        if self._onlyPauseShift and pauseTime is None:
            return 0.0
        return (
            currentTime
            if pauseTime is None
            else currentTime - pauseTime.timestamp()
        )


_DateScaleTypeCheck: type[Scale[DateTime[ZoneInfo], float, float]] = DateScale


@dataclass(frozen=True)
class DateTimeDriver:
    """
    Driver based on ZoneInfo-aware datetimes.

    @ivar driver: the L{TimeDriver} that this L{DateTimeDriver} is layered on
        top of, one which represents time as a POSIX timestamp as a L{float}.

    @ivar zone: the default timezone of this L{DateTimeDriver}, the one which
        will be used for results from L{now() <DateTimeDriver.now>}.  Note that
        L{reschedule(...) <DateTimeDriver.reschedule>} will still take inputs
        in any zone; this is just the default zone for outputs.
    """

    driver: TimeDriver[float]
    zone: ZoneInfo = ZoneInfo("Etc/UTC")

    def unschedule(self) -> None:
        "Implementation of L{TimeDriver.unschedule}"
        self.driver.unschedule()

    def reschedule(
        self, newTime: DateTime[ZoneInfo], work: Callable[[], None]
    ) -> None:
        "Implementation of L{TimeDriver.reschedule}"
        self.driver.reschedule(newTime.timestamp(), work)

    def now(self) -> DateTime[ZoneInfo]:
        "Implementation of L{TimeDriver.now}"
        timestamp = self.driver.now()
        return DateTime.fromtimestamp(timestamp, self.zone)


_DriverTypeCheck: type[TimeDriver[DateTime[ZoneInfo]]] = DateTimeDriver



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import datetime, timedelta
from json import dumps, loads
from pathlib import Path
from tempfile import mkdtemp
from typing import Any, Callable, Type
from unittest import TestCase
from zoneinfo import ZoneInfo

from datetype import DateTime, aware

from ..boundaries import (
    Cancellable,
    RecurrenceRule,
    ScheduledCall,
    ScheduledState,
    SomeScheduledCall,
    TimeDriver,
)
from ..drivers.datetimes import DateTimeDriver
from ..drivers.memory import MemoryDriver
from ..persistent.jsonable import (
    JSONableCallable,
    JSONableInstance,
    JSONableRepeatable,
    JSONableScheduler,
    JSONObject,
    JSONRegistry,
    LoadProcess,
    MissingPersistentCall,
    schedulerAtPath,
)
from ..repeat.rules.datetimes import EachYear, daily


@dataclass
class RegInfo:
    madeCalls: list[str]
    identityMap: dict[str, Any] = field(default_factory=dict)
    lookupLater: list[LaterStopper] = field(default_factory=list)


registry = JSONRegistry[RegInfo]()
emptyRegistry = JSONRegistry[RegInfo]()
PT = ZoneInfo(key="America/Los_Angeles")

globalCalls = []


@registry.function
def call1() -> None:
    globalCalls.append("hello")


@registry.function
def call2() -> None:
    globalCalls.append("goodbye")


@registry.repeatFunction
def repeatable(steps: int, scheduled: SomeScheduledCall) -> None:
    globalCalls.append(f"repeatable {steps}")


@dataclass
class InstanceWithMethods:
    value: str
    info: RegInfo
    callCount: int = 0
    stoppers: list[Cancellable] = field(default_factory=list)

    @classmethod
    def typeCodeForJSON(self) -> str:
        return "instanceWithMethods"

    @classmethod
    def fromJSON(
        cls, load: LoadProcess[RegInfo], json: JSONObject
    ) -> InstanceWithMethods:
        key = json["identity"]
        if key in load.bootstrap.identityMap:
            load.bootstrap.madeCalls.append(
                f"InstanceWithMethods.fromJSON: {json['value']} (cached)"
            )
            self: InstanceWithMethods = load.bootstrap.identityMap[key]
            return self
        load.bootstrap.madeCalls.append(
            f"InstanceWithMethods.fromJSON: {json['value']}"
        )
        new = cls(json["value"], load.bootstrap)
        load.bootstrap.identityMap[key] = new
        return new

    def toJSON(self, registry: JSONRegistry[RegInfo]) -> dict[str, object]:
        return {
            "value": self.value,
            "identity": id(self),
        }

    @registry.method
    def method1(self) -> None:
        self.info.madeCalls.append(f"{self.value}/method1")

    @registry.method
    def method2(self) -> None:
        self.info.madeCalls.append(f"{self.value}/method2")

    @registry.repeatMethod
    def repeatMethod(self, steps: int, scheduled: SomeScheduledCall) -> None:
        self.callCount += 1
        self.stoppers.append(scheduled)
        self.info.madeCalls.append(
            f"repeatMethod {steps} {self.value=} {self.callCount=}"
        )

    @registry.repeatMethod
    def repeatMethodDTZIL(
        self, steps: list[DateTime[ZoneInfo]], scheduled: SomeScheduledCall
    ) -> None:
        self.callCount += 1
        self.stoppers.append(scheduled)
        self.info.madeCalls.append(
            f"repeatMethod {steps} {self.value=} {self.callCount=}"
        )


Handle = ScheduledCall[DateTime[ZoneInfo], JSONableCallable[RegInfo], int]


@dataclass
class LaterStopper:
    handle: Handle

    @registry.method
    def stop(self) -> None:
        self.handle.cancel()

    @classmethod
    def typeCodeForJSON(self) -> str:
        return "later-stopper"

    def toJSON(self, registry: JSONRegistry[RegInfo]) -> dict[str, object]:
        return {
            "value": registry.saveScheduledCall(self.handle),
        }

    @classmethod
    def fromJSON(
        cls, load: LoadProcess[RegInfo], json: JSONObject
    ) -> LaterStopper:
        new = cls(load.loadScheduledCall(json["value"]))
        load.bootstrap.lookupLater.append(new)
        return new


@dataclass
class Stoppable:
    runcall: Handle | None = None
    stopcall: Handle | None = None
    ran: bool = False

    def scheduleme(self, scheduler: JSONableScheduler[RegInfo]) -> None:
        """
        Schedule 'runme' to run 2 seconds in the future, but 'stopme' to run 1
        second in the future.
        """
        now = scheduler.now()
        self.runcall = scheduler.callAt(now + timedelta(seconds=2), self.runme)
        self.stopcall = scheduler.callAt(
            now + timedelta(seconds=1), self.stopme
        )

    @classmethod
    def typeCodeForJSON(self) -> str:
        return "stoppable"

    def toJSON(self, registry: JSONRegistry[RegInfo]) -> dict[str, object]:
        def save(it: Handle | None) -> object:
            return registry.saveScheduledCall(it) if it is not None else it

        return {
            "runcall": save(self.runcall),
            "stopcall": save(self.stopcall),
            "ran": self.ran,
            "id": id(self),
        }

    @classmethod
    def fromJSON(
        cls, load: LoadProcess[RegInfo], json: JSONObject
    ) -> Stoppable:
        ckey = json["id"]
        if ckey in load.bootstrap.identityMap:
            result: Stoppable = load.bootstrap.identityMap[ckey]
            return result

        def get(
            name: str,
        ) -> Handle | None:
            it = json[name]
            assert it is not None
            loaded = load.loadScheduledCall(it)
            assert loaded.state is ScheduledState.pending
            return loaded

        self = cls(
            runcall=get("runcall"),
            stopcall=get("stopcall"),
            ran=json["ran"],
        )
        # leave it there for the test to pick up
        load.bootstrap.identityMap[ckey] = self
        return self

    @registry.method
    def stopme(self) -> None:
        assert self.runcall is not None
        self.stopcall = None
        self.runcall.cancel()
        self.runcall = None

    @registry.method
    def runme(self) -> None:
        self.ran = True
        self.runcall = None


stp: Type[JSONableInstance[RegInfo]] = Stoppable


def jsonScheduler(
    driver: TimeDriver[float],
) -> tuple[JSONableScheduler[RegInfo], Callable[[], JSONObject]]:
    return registry.createScheduler(DateTimeDriver(driver))


class PersistentSchedulerTests(TestCase):
    def tearDown(self) -> None:
        del globalCalls[:]

    def test_scheduleRunSaveRun(self) -> None:
        """
        Test scheduling module-level functions and instance methods.
        """
        memoryDriver = MemoryDriver()
        scheduler, saver = jsonScheduler(memoryDriver)
        dt = aware(
            datetime(2023, 7, 21, 1, 1, 1, tzinfo=PT),
            ZoneInfo,
        )
        dt2 = aware(
            datetime(2023, 7, 22, 1, 1, 1, tzinfo=PT),
            ZoneInfo,
        )
        ri0 = RegInfo([])
        iwm = InstanceWithMethods("test_scheduleRunSaveRun-value", ri0)
        scheduler.callAt(dt, call1)
        scheduler.callAt(dt, iwm.method1)
        scheduler.callAt(dt2, call2)
        scheduler.callAt(dt2, iwm.method2)
        memoryDriver.advance(dt.timestamp() + 1)
        self.assertEqual(globalCalls, ["hello"])
        del globalCalls[:]
        saved = saver()
        memory2 = MemoryDriver()
        ri = RegInfo([])
        registry.loadScheduler(DateTimeDriver(memory2), saved, ri)
        memory2.advance(dt2.timestamp() + 1)
        self.assertEqual(globalCalls, ["goodbye"])
        self.assertEqual(
            ri0.madeCalls, ["test_scheduleRunSaveRun-value/method1"]
        )
        self.assertEqual(
            ri.madeCalls,
            [
                "InstanceWithMethods.fromJSON: test_scheduleRunSaveRun-value",
                "test_scheduleRunSaveRun-value/method2",
            ],
        )

    def test_persistCancellers(self) -> None:
        """
        scheduled instance methods ought to be able to save handles to other
        instances and stuff
        """
        memoryDriver = MemoryDriver()
        scheduler, saver = jsonScheduler(memoryDriver)
        dt = aware(datetime(2023, 7, 21, 1, 1, 1, tzinfo=PT), ZoneInfo)
        memoryDriver.advance(dt.timestamp() + 1)
        s = Stoppable()
        self.assertEqual(s.ran, False)
        s.runme()
        self.assertEqual(s.ran, True)

        s = Stoppable()
        s.scheduleme(scheduler)
        assert s.runcall is not None
        assert s.stopcall is not None
        s2 = LaterStopper(s.runcall)
        s3 = LaterStopper(s.stopcall)
        s4 = LaterStopper(s.runcall)
        s5 = LaterStopper(s.stopcall)
        scheduler.callAt(
            aware(datetime(2029, 1, 1, tzinfo=PT), ZoneInfo), s2.stop
        )
        scheduler.callAt(
            aware(datetime(2029, 1, 2, tzinfo=PT), ZoneInfo), s3.stop
        )
        scheduler.callAt(
            aware(datetime(2029, 1, 3, tzinfo=PT), ZoneInfo), s4.stop
        )
        last = scheduler.callAt(
            aware(datetime(2029, 1, 4, tzinfo=PT), ZoneInfo), s5.stop
        )
        scheduler.callAt(
            aware(datetime(2029, 1, 5, tzinfo=PT), ZoneInfo),
            LaterStopper(last).stop,
        )
        jsonobj = dumps(saver())
        saved = loads(jsonobj)
        memory2 = MemoryDriver()
        ri = RegInfo([])
        registry.loadScheduler(DateTimeDriver(memory2), saved, ri)
        [run1, stop1, run2, stop2, loadedLast] = ri.lookupLater
        self.assertEqual(run1.handle.state, ScheduledState.pending)
        [(name, loadedStoppable)] = ri.identityMap.items()
        assert isinstance(loadedStoppable, Stoppable)
        self.assertEqual(loadedStoppable.ran, False)
        rc = loadedStoppable.runcall
        sc = loadedStoppable.stopcall
        assert rc is not None
        assert sc is not None
        self.assertEqual(rc.state, ScheduledState.pending)
        self.assertEqual(
            rc.when,
            datetime(2023, 7, 21, 8, 1, 4, tzinfo=ZoneInfo(key="Etc/UTC")),
        )
        self.assertIsNot(rc.what, None)
        self.assertEqual(rc.id, 0)
        memory2.advance(dt.timestamp() + 4.0)
        self.assertEqual(loadedStoppable.ran, False)
        self.assertIs(loadedStoppable.runcall, None)
        self.assertEqual(rc.state, ScheduledState.cancelled)
        self.assertEqual(loadedLast.handle.state, ScheduledState.pending)
        loadedLast.stop()
        self.assertEqual(loadedLast.handle.state, ScheduledState.cancelled)

        # These are loaded reentrantly and cannot be identical, but they map
        # their IDs and are the same
        self.assertEqual(rc.id, run1.handle.id)
        self.assertEqual(sc.id, stop1.handle.id)

        # all references not loaded reentrantly are exactly identical
        self.assertIs(run1.handle, run2.handle)
        self.assertEqual(run1.handle.state, ScheduledState.cancelled)
        self.assertIs(stop1.handle, stop2.handle)

    def test_schedulerAtPath(self) -> None:
        ri0 = RegInfo([])
        iwm = InstanceWithMethods("A", ri0)
        mem = MemoryDriver()
        p = Path(mkdtemp()) / "scheduler.json"
        ts = datetime(2024, 2, 1, tzinfo=ZoneInfo("Etc/UTC")).timestamp()
        mem.advance(ts)
        aw = aware(datetime(2024, 2, 2, tzinfo=ZoneInfo("Etc/UTC")), ZoneInfo)
        with schedulerAtPath(registry, DateTimeDriver(mem), p, ri0) as sched1:
            sched1.callAt(aw, iwm.method1)
        ri1 = RegInfo([])
        mem2 = MemoryDriver()
        with schedulerAtPath(registry, DateTimeDriver(mem2), p, ri1):
            mem2.advance()
        self.assertEqual(
            ri1.madeCalls, ["InstanceWithMethods.fromJSON: A", "A/method1"]
        )

    def test_noSuchCallID(self) -> None:
        mem = MemoryDriver()
        with self.assertRaises(MissingPersistentCall) as raised:
            registry.loadScheduler(
                DateTimeDriver(mem),
                {
                    "scheduledCalls": [
                        {
                            "when": "2023-07-21T08:01:03",
                            "tz": "Etc/UTC",
                            "what": {
                                "type": "stoppable.stopme",
                                "data": {
                                    "runcall": {"id": 7},
                                    "stopcall": {"id": 2},
                                    "ran": False,
                                    "id": 4411099664,
                                },
                            },
                            "id": 2,
                        }
                    ],
                    "counter": "2",
                },
                RegInfo([]),
            )
        self.assertEqual(raised.exception.args[0], 7)

    def test_idling(self) -> None:
        memoryDriver = MemoryDriver()
        scheduler, saver = jsonScheduler(memoryDriver)
        dt = aware(
            datetime(2023, 7, 21, 1, 1, 1, tzinfo=PT),
            ZoneInfo,
        )
        handle = scheduler.callAt(dt, call1)
        self.assertEqual(memoryDriver.isScheduled(), True)
        handle.cancel()
        self.assertEqual(memoryDriver.isScheduled(), False)
        memoryDriver.advance(dt.timestamp() + 1)
        self.assertEqual(globalCalls, [])

    def test_emptyScheduler(self) -> None:
        memory = MemoryDriver()
        registry.loadScheduler(
            DateTimeDriver(memory), {"scheduledCalls": []}, RegInfo([])
        )
        self.assertEqual(memory.isScheduled(), False)

    def test_repeatable(self) -> None:
        dt = aware(
            datetime(2023, 7, 21, 1, 1, 1, tzinfo=PT),
            ZoneInfo,
        )
        memoryDriver = MemoryDriver()
        memoryDriver.advance(dt.timestamp())
        scheduler, saver = jsonScheduler(memoryDriver)
        registry.repeatedly(scheduler, daily, repeatable, dt)
        self.assertEqual(globalCalls, ["repeatable 1"])
        del globalCalls[:]

        def days(n: int) -> float:
            return 60 * 60 * 24 * n

        memoryDriver.advance(days(3))
        self.assertEqual(globalCalls, ["repeatable 3"])
        del globalCalls[:]

        newInfo = RegInfo([])
        mem2 = MemoryDriver()
        mem2.advance(dt.timestamp())
        mem2.advance(days(7))
        self.assertEqual(mem2.isScheduled(), False)
        registry.loadScheduler(
            DateTimeDriver(mem2), loads(dumps(saver())), newInfo
        )
        self.assertEqual(mem2.isScheduled(), True)
        amount = mem2.advance()
        assert amount is not None
        self.assertLess(amount, 0.0001)
        self.assertEqual(globalCalls, ["repeatable 4"])

    def test_repeatEachYear(self) -> None:
        memoryDriver = MemoryDriver()
        dt = aware(
            datetime(2023, 7, 21, 1, 1, 1, tzinfo=PT),
            ZoneInfo,
        )
        scheduler: JSONableScheduler[RegInfo]
        scheduler, saver = jsonScheduler(memoryDriver)
        ri = RegInfo([])
        iwm = InstanceWithMethods("test_repeatEachYear", ri)
        rrule: RecurrenceRule[DateTime[ZoneInfo], list[DateTime[ZoneInfo]]] = (
            EachYear(2)
        )
        repeatMethod: JSONableRepeatable[RegInfo, list[DateTime[ZoneInfo]]] = (
            iwm.repeatMethodDTZIL
        )
        registry.repeatedly(scheduler, rrule, repeatMethod, dt)
        newInfo = RegInfo([])
        mem2 = MemoryDriver()
        mem2.advance(dt.timestamp())
        registry.loadScheduler(
            DateTimeDriver(mem2), loads(dumps(saver())), newInfo
        )
        mem2.advance(timedelta(days=365 * 4).total_seconds())
        LA = "zoneinfo.ZoneInfo(key='America/Los_Angeles')"
        expectedCalls = [
            "InstanceWithMethods.fromJSON: test_repeatEachYear",
            "repeatMethod ["
            f"datetime.datetime(2023, 7, 21, 1, 1, 1, tzinfo={LA}), "
            f"datetime.datetime(2025, 7, 21, 1, 1, 1, tzinfo={LA})"
            "] self.value='test_repeatEachYear' self.callCount=1",
        ]
        self.assertEqual(newInfo.madeCalls, expectedCalls)

    def test_repeatLoadError(self) -> None:
        dt = aware(
            datetime(2023, 7, 21, 1, 1, 1, tzinfo=PT),
            ZoneInfo,
        )
        memoryDriver = MemoryDriver()
        memoryDriver.advance(dt.timestamp())
        oneCall = {
            "when": "2023-07-22T08:01:01",
            "tz": "Etc/UTC",
            "what": {
                "type": "fritter:repetition.repeat",
                "data": {
                    "ts": "2023-07-22T08:01:01",
                    "tz": "Etc/UTC",
                    "rule": {
                        "type": "incorrect rule type",
                        "data": {"delta": (1, 0, 0)},
                    },
                    "callable": {
                        "type": "instanceWithMethods.repeatMethod",
                        "data": {
                            "value": "sample",
                            "identity": 4335201296,
                        },
                    },
                },
            },
            "called": False,
            "cancelled": False,
            "id": 1,
        }
        with self.assertRaises(KeyError) as ke:
            registry.loadScheduler(
                DateTimeDriver(memoryDriver),
                {
                    "scheduledCalls": [oneCall],
                    "counter": "1",
                },
                RegInfo([]),
            )
        self.assertEqual(
            str(ke.exception),
            repr("cannot interpret rule type code 'incorrect rule type'"),
        )

    def test_repeatableMethod(self) -> None:
        dt = aware(
            datetime(2023, 7, 21, 1, 1, 1, tzinfo=PT),
            ZoneInfo,
        )
        memoryDriver = MemoryDriver()
        memoryDriver.advance(dt.timestamp())
        scheduler, saver = jsonScheduler(memoryDriver)
        info = RegInfo([])
        inst = InstanceWithMethods("sample", info)
        method = inst.repeatMethod
        shared = InstanceWithMethods("shared", info)
        registry.repeatedly(scheduler, daily, method)
        registry.repeatedly(scheduler, daily, shared.repeatMethod)
        registry.repeatedly(scheduler, daily, shared.repeatMethod)
        self.assertEqual(
            info.madeCalls,
            [
                "repeatMethod 1 self.value='sample' self.callCount=1",
                "repeatMethod 1 self.value='shared' self.callCount=1",
                "repeatMethod 1 self.value='shared' self.callCount=2",
            ],
        )
        del info.madeCalls[:]

        def days(n: int) -> float:
            return 60 * 60 * 24 * n

        memoryDriver.advance(days(3))
        expected = [
            "repeatMethod 3 self.value='sample' self.callCount=2",
            "repeatMethod 3 self.value='shared' self.callCount=3",
            "repeatMethod 3 self.value='shared' self.callCount=4",
        ]

        self.assertEqual(info.madeCalls, expected)

        newInfo = RegInfo([])
        newNewInfo = RegInfo([])

        def atTimeDriver() -> MemoryDriver:
            x = MemoryDriver()
            x.advance(dt.timestamp())
            x.advance(days(7))
            return x

        mem2 = atTimeDriver()
        mem3 = atTimeDriver()

        self.assertEqual(mem2.isScheduled(), False)
        persistent = dumps(saver())
        loadedScheduler, saver2 = registry.loadScheduler(
            DateTimeDriver(mem2), loads(persistent), newInfo
        )
        repersistent = dumps(saver2())
        registry.loadScheduler(
            DateTimeDriver(mem3), loads(repersistent), newNewInfo
        )
        loaded = loads(persistent)
        with self.assertRaises(KeyError):
            # TODO: allow for better error handling that doesn't just blow up
            # on the type code lookup failure
            emptyRegistry.loadScheduler(
                DateTimeDriver(MemoryDriver()), loaded, newInfo
            )
        self.assertEqual(mem2.isScheduled(), True)
        mem2.advance()
        expectedCalls = [
            "InstanceWithMethods.fromJSON: sample",
            "InstanceWithMethods.fromJSON: shared",
            "InstanceWithMethods.fromJSON: shared (cached)",
            "repeatMethod 4 self.value='sample' self.callCount=1",
            "repeatMethod 4 self.value='shared' self.callCount=1",
            "repeatMethod 4 self.value='shared' self.callCount=2",
        ]
        self.assertEqual(newInfo.madeCalls, expectedCalls)
        self.assertEqual(mem3.isScheduled(), True)
        mem3.advance()
        self.assertEqual(newNewInfo.madeCalls, expectedCalls)

        # round trip:



from __future__ import annotations

from asyncio.events import new_event_loop
from asyncio.exceptions import CancelledError, InvalidStateError
from contextvars import Context
from dataclasses import dataclass
from sys import version_info
from typing import Any, Callable
from unittest import TestCase

from twisted.internet.task import Clock

from ..boundaries import Cancellable
from ..drivers.asyncio import AsyncioAsyncDriver, AsyncioTimeDriver, scheduler

if version_info >= (3, 11):
    from typing import Unpack, TypeVarTuple
else:
    from typing_extensions import Unpack, TypeVarTuple

_Ts = TypeVarTuple("_Ts")


@dataclass
class AsyncioClock:
    """
    Minimal translation of Twisted interface into Asyncio interface.
    """

    clock: Clock

    def call_at(
        self,
        when: float,
        callback: Callable[[Unpack[_Ts]], object],
        *args: Unpack[_Ts],
        context: Context | None = None,
    ) -> Cancellable:
        assert context is None, "context not yet supported"
        return self.clock.callLater(
            when - self.clock.seconds(),
            callback,
            *args,
        )

    def time(self) -> float:
        return self.clock.seconds()


class AsyncDriverTests(TestCase):
    def setUp(self) -> None:
        self.called = 0
        self.loop = new_event_loop()

    def tearDown(self) -> None:
        self.loop.close()

    def call(self) -> None:
        self.called += 1

    def test_complete(self) -> None:
        driver = AsyncioAsyncDriver(self.loop)

        f = driver.newWithCancel(self.call)
        with self.assertRaises(InvalidStateError):
            f.result()

        async def a() -> None:
            driver.complete(f)
            await f

        self.loop.run_until_complete(a())
        self.assertEqual(self.called, 0)
        self.assertIsNone(f.result())

    def test_cancel(self) -> None:
        driver = AsyncioAsyncDriver(self.loop)
        f = driver.newWithCancel(self.call)
        with self.assertRaises(InvalidStateError):
            f.result()
        self.assertEqual(self.called, 0)

        async def a() -> None:
            f.cancel()
            with self.assertRaises(CancelledError):
                await f

        self.loop.run_until_complete(a())
        self.assertEqual(self.called, 1)
        with self.assertRaises(CancelledError):
            f.result()

    def test_runAsync(self) -> None:
        driver = AsyncioAsyncDriver(self.loop)

        # Set up a bunch of futures to ensure tasks are run independently
        # because runAsync doesn't return an awaitable.

        f1 = driver.newWithCancel(self.call)
        f2 = driver.newWithCancel(self.call)
        f3 = driver.newWithCancel(self.call)
        f4 = driver.newWithCancel(self.call)
        f5 = driver.newWithCancel(self.call)
        f6 = driver.newWithCancel(self.call)
        e1 = driver.newWithCancel(self.call)
        e2 = driver.newWithCancel(self.call)
        concurrent = []

        async def c1() -> None:
            concurrent.append("c1")
            await f1
            f3.set_result(None)
            await e1
            concurrent.append("c3")
            f5.set_result(None)

        async def c2() -> None:
            concurrent.append("c2")
            await f2
            f4.set_result(None)
            await e2
            concurrent.append("c4")
            f6.set_result(None)

        async def t() -> None:
            f1.set_result(None)
            f2.set_result(None)
            await f3
            await f4
            self.assertEqual(concurrent, ["c1", "c2"])
            e1.set_result(None)
            e2.set_result(None)
            await f5
            await f6
            self.assertEqual(concurrent, ["c1", "c2", "c3", "c4"])

        driver.runAsync(c1())
        driver.runAsync(c2())

        self.loop.run_until_complete(t())
        self.assertEqual(self.called, 0)


class TimeDriverTests(TestCase):
    def setUp(self) -> None:
        self.called = 0

    def call(self) -> None:
        self.called += 1

    def test_schedule(self) -> None:
        driver = AsyncioTimeDriver(AsyncioClock(clock := Clock()))
        driver.reschedule(3.0, self.call)
        self.assertEqual(self.called, 0)
        clock.advance(4.0)
        self.assertEqual(self.called, 1)

    def test_now(self) -> None:
        driver = AsyncioTimeDriver(AsyncioClock(clock := Clock()))
        clock.advance(7.2)
        self.assertEqual(driver.now(), 7.2)

    def test_reschedule(self) -> None:
        driver = AsyncioTimeDriver(AsyncioClock(clock := Clock()))
        driver.reschedule(3.0, self.call)
        self.assertEqual(self.called, 0)
        driver.reschedule(5.0, self.call)
        clock.advance(4.0)
        self.assertEqual(self.called, 0)
        clock.advance(1.0)
        self.assertEqual(self.called, 1)

    def test_unschedule(self) -> None:
        driver = AsyncioTimeDriver(AsyncioClock(clock := Clock()))
        driver.reschedule(3.0, self.call)
        self.assertEqual(self.called, 0)
        driver.unschedule()
        clock.advance(4.0)
        self.assertEqual(self.called, 0)
        driver.unschedule()  # safe no-op

    def test_scheduler(self) -> None:
        sched = scheduler(AsyncioClock(clock := Clock()))
        stuff = []

        def hello() -> None:
            stuff.append("hello")

        sched.callAt(50, hello)
        self.assertEqual(stuff, [])
        clock.advance(60)
        self.assertEqual(stuff, ["hello"])

    def test_schedulerDefaults(self) -> None:
        sched: Any = scheduler()
        self.assertIsInstance(sched.driver, AsyncioTimeDriver)



"""
Implementation of L{TimeDriver} and L{AsyncDriver} in terms of U{Twisted
<https://twisted.org/>}'s APIs, L{IReactorTime} and L{Deferred}.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Coroutine, Optional

from twisted.internet.defer import Deferred
from twisted.internet.interfaces import IDelayedCall, IReactorTime
from twisted.logger import Logger

from ..boundaries import (
    AsyncDriver,
    PhysicalScheduler,
    PriorityQueue,
    TimeDriver,
)
from ..heap import Heap
from ..scheduler import ConcreteScheduledCall, schedulerFromDriver

log = Logger()


@dataclass
class TwistedTimeDriver:
    """
    Instantiate a L{TwistedTimeDriver} with an L{IReactorTime}; for example::

        from twisted.internet.task import react

        async def main(reactor):
            driver = TwistedTimeDriver(reactor)
            ...

        task.react(main)
    """

    _reactor: IReactorTime
    _call: Optional[IDelayedCall] = None

    def reschedule(self, desiredTime: float, work: Callable[[], None]) -> None:
        """"""

        def _() -> None:
            self._call = None
            work()

        if self._call is not None:
            self._call.cancel()
        self._call = self._reactor.callLater(
            max(0, desiredTime - self.now()), _
        )

    def unschedule(self) -> None:
        if self._call is not None:
            self._call.cancel()
            self._call = None

    def now(self) -> float:
        return self._reactor.seconds()


_TimeDriverCheck: type[TimeDriver[float]] = TwistedTimeDriver


@dataclass
class TwistedAsyncDriver:
    """
    Driver for Deferred-flavored awaitables.
    """

    def newWithCancel(self, cancel: Callable[[], None]) -> Deferred[None]:
        """
        Create a new future-ish object with the given callback to execute when
        cancelled.
        """
        return Deferred(lambda d: cancel())

    def complete(self, asyncObj: Deferred[None]) -> None:
        """
        The asynchronous operation completed successfully.
        """
        asyncObj.callback(None)

    def runAsync(
        self, coroutine: Coroutine[Deferred[None], Any, Any]
    ) -> Deferred[None]:
        return Deferred.fromCoroutine(coroutine)


_AsyncDriverCheck: type[AsyncDriver[Deferred[None]]] = TwistedAsyncDriver


def scheduler(
    reactor: IReactorTime | None = None,
    queue: (
        PriorityQueue[ConcreteScheduledCall[float, Callable[[], None], int]]
        | None
    ) = None,
) -> PhysicalScheduler:
    """
    Create a scheduler that uses Twisted.
    """
    if reactor is None:
        from twisted.internet import reactor  # type:ignore[assignment]

        assert reactor is not None
    return schedulerFromDriver(
        TwistedTimeDriver(reactor),
        queue=queue if queue is not None else Heap(),
    )



# -*- test-case-name: fritter.test.test_asyncio -*-

"""
Implementation of L{TimeDriver} and L{AsyncDriver} for L{asyncio}.
"""

from __future__ import annotations

from asyncio import AbstractEventLoop, Future, get_event_loop
from contextvars import Context
from dataclasses import dataclass, field
from typing import Any, Callable, Coroutine, Protocol

from ..boundaries import (
    AsyncDriver,
    Cancellable,
    PriorityQueue,
    Scheduler,
    TimeDriver,
)
from ..heap import Heap
from ..scheduler import ConcreteScheduledCall, schedulerFromDriver
from sys import version_info

if version_info >= (3, 11):
    from typing import Unpack, TypeVarTuple
else:
    from typing_extensions import Unpack, TypeVarTuple

_Ts = TypeVarTuple("_Ts")


class LoopTimeInterface(Protocol):
    """
    Describe the portions of L{AbstractEventLoop} used by L{AsyncioTimeDriver}.
    """

    def call_at(
        self,
        when: float,
        callback: Callable[[Unpack[_Ts]], object],
        *args: Unpack[_Ts],
        context: Context | None = None,
    ) -> Cancellable:
        """
        @see: L{AbstractEventLoop.call_at <asyncio.loop.call_at>}
        """

    def time(self) -> float:
        """
        @see: L{AbstractEventLoop.time <asyncio.loop.time>}
        """


@dataclass
class AsyncioTimeDriver:
    """
    An implementation of L{TimeDriver} using an L{asyncio} event loop.
    """

    _loop: LoopTimeInterface = field(default_factory=get_event_loop)
    _call: Cancellable | None = None

    def reschedule(self, desiredTime: float, work: Callable[[], None]) -> None:
        "Implementation of L{TimeDriver.reschedule}"

        def _() -> None:
            self._call = None
            work()

        if self._call is not None:
            self._call.cancel()
        self._call = self._loop.call_at(max(0, desiredTime), _)

    def unschedule(self) -> None:
        "Implementation of L{TimeDriver.unschedule}"
        if self._call is not None:
            self._call.cancel()
            self._call = None

    def now(self) -> float:
        "Implementation of L{TimeDriver.now}"
        return self._loop.time()


_TimeDriverCheck: type[TimeDriver[float]] = AsyncioTimeDriver


@dataclass
class AsyncioAsyncDriver:
    """
    Driver for L{Future}-flavored awaitables.

    @see: L{fritter.repeat.Async}

    @see: L{TimeDriver}
    """

    _loop: AbstractEventLoop = field(default_factory=get_event_loop)

    def newWithCancel(self, cancel: Callable[[], None]) -> Future[None]:
        """
        Create a new L{Future} with the given callback to execute when
        cancelled.
        """
        f = Future[None](loop=self._loop)

        @f.add_done_callback
        def done(future: Future[None]) -> None:
            if f.cancelled():
                cancel()

        return f

    def complete(self, asyncObj: Future[None]) -> None:
        "The asynchronous operation completed successfully."
        asyncObj.set_result(None)

    def runAsync(
        self, coroutine: Coroutine[Future[None], Any, Any]
    ) -> Future[None]:
        """
        Run the given task on the event loop with L{asyncio.create_task}.
        """
        return self._loop.create_task(coroutine)


_AsyncioDriverCheck: type[AsyncDriver[Future[None]]] = AsyncioAsyncDriver


def scheduler(
    loop: LoopTimeInterface | None = None,
    queue: (
        PriorityQueue[ConcreteScheduledCall[float, Callable[[], None], int]]
        | None
    ) = None,
) -> Scheduler[float, Callable[[], None], int]:
    """
    Create a scheduler that uses Asyncio.
    """
    return schedulerFromDriver(
        AsyncioTimeDriver(loop if loop is not None else get_event_loop()),
        queue=queue if queue is not None else Heap(),
    )
